"""
Utilities to implement functional retry policies for coroutines.
"""

import logging
import time
from functools import wraps
from typing import Awaitable, Callable, Sequence, Type, TypeVar

from typing_extensions import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


logger = logging.getLogger(__name__)


class RetryLimitExceeded(Exception):
    """Function raised more exceptions than allowed."""


def retry_policy(
    min_retry_interval_seconds: float = 1,
    max_retry_interval_seconds: float = 60,
    max_retry_count: int = 10,
    retryable_exceptions: Sequence[Type[Exception]] = (Exception,),
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    """Decorate function to execute in managed retry environment."""

    def coroutine_with_retry_policy(
        coroutine: Callable[P, Awaitable[T]]
    ) -> Callable[P, Awaitable[T]]:
        @wraps(coroutine)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            retry_count = 0
            retry_interval_seconds = min_retry_interval_seconds

            while True:
                try:
                    return await coroutine(*args, **kwargs)
                except tuple(retryable_exceptions) as e:
                    if retry_count >= max_retry_count:
                        # Construct dynamically created exception which subclasses from
                        # RetryLimitExceeded and the exception raised from the callable
                        # to preserve exception handling behavior of the caller.
                        raise type("RetryFailed", (RetryLimitExceeded, type(e)), {})

                    logger.warning("Retrying exception: %s", repr(e))
                    time.sleep(retry_interval_seconds)

                    retry_count += 1
                    retry_interval_seconds = min(
                        retry_interval_seconds * 2, max_retry_interval_seconds
                    )

        return wrapper

    return coroutine_with_retry_policy



import asyncio
from abc import ABC
from datetime import datetime
from typing import Any, Dict, List, Optional, TypedDict

from pydantic import validate_arguments

from ..cache import cache
from ..core import BaseClient
from ..encoders import encode_query_params
from ..errors import RequestError, ServerError, UserNotFoundError
from ..retry import retry_policy

PATH = "/users"
USER_PATH = "/users/{user_uuid}"

User = Dict[str, Any]


class UsersResponse(TypedDict):
    page: int
    num_pages: int
    users: List[User]


class BaseUsersResource(ABC):
    def __init__(self, client: BaseClient):
        self._client = client

    @retry_policy(
        max_retry_count=4,
        max_retry_interval_seconds=10,
        retryable_exceptions=[ServerError],
    )
    async def _get_page(
        self,
        created_after: Optional[datetime] = None,
        created_before: Optional[datetime] = None,
        search: Optional[str] = None,
        phone_number: Optional[str] = None,
        device_id: Optional[str] = None,
        external_id: Optional[str] = None,
        page: int = 1,
    ) -> UsersResponse:
        async with self._client.session() as session:
            response = await session.get(
                PATH,
                params=encode_query_params(
                    created_after=created_after,
                    created_before=created_before,
                    search=search,
                    phone_number=phone_number,
                    device_id=device_id,
                    external_id=external_id,
                    page=page,
                ),
            )

        if response.status_code // 100 == 5:
            raise ServerError(response.text)

        if response.status_code // 100 == 4:
            raise RequestError(response.text)

        users_response: UsersResponse = response.json()
        return users_response

    @validate_arguments
    @cache
    async def _get(
        self,
        created_after: Optional[datetime] = None,
        created_before: Optional[datetime] = None,
        search: Optional[str] = None,
        phone_number: Optional[str] = None,
        device_id: Optional[str] = None,
        external_id: Optional[str] = None,
    ) -> List[User]:
        response = await self._get_page(
            created_after=created_after,
            created_before=created_before,
            search=search,
            device_id=device_id,
            external_id=external_id,
            phone_number=phone_number,
        )
        max_pages = response["num_pages"]
        if max_pages > 1:
            coroutines = [
                self._get_page(
                    created_after=created_after,
                    created_before=created_before,
                    search=search,
                    device_id=device_id,
                    external_id=external_id,
                    page=page + 2,
                )
                for page in range(max_pages - 1)
            ]
            response_pages = await asyncio.gather(*coroutines)
            responses = (response, *response_pages)
        else:
            responses = (response,)

        return [user for response in responses for user in response["users"]]

    @retry_policy(
        max_retry_count=4,
        max_retry_interval_seconds=10,
        retryable_exceptions=[ServerError],
    )
    @validate_arguments
    @cache
    async def _get_user(
        self,
        user_uuid: str,
    ) -> List[User]:
        async with self._client.session() as session:
            response = await session.get(USER_PATH.format(user_uuid=user_uuid))

        if response.status_code // 100 == 5:
            raise ServerError(response.text)

        if response.status_code == 404:
            raise UserNotFoundError(response.text)

        if response.status_code // 100 == 4:
            raise RequestError(response.text)

        user: User = response.json()
        return [user]


class AsyncUsersResource(BaseUsersResource):
    async def get(
        self,
        *,
        created_after: Optional[datetime] = None,
        created_before: Optional[datetime] = None,
        search: Optional[str] = None,
        user_uuid: Optional[str] = None,
        phone_number: Optional[str] = None,
        device_id: Optional[str] = None,
        external_id: Optional[str] = None,
    ) -> List[User]:
        if user_uuid:
            return await self._get_user(
                user_uuid=user_uuid,
            )

        return await self._get(
            created_after=created_after,
            created_before=created_before,
            search=search,
            phone_number=phone_number,
            device_id=device_id,
            external_id=external_id,
        )


class SyncUsersResource(BaseUsersResource):
    def get(
        self,
        *,
        created_after: Optional[datetime] = None,
        created_before: Optional[datetime] = None,
        search: Optional[str] = None,
        user_uuid: Optional[str] = None,
        phone_number: Optional[str] = None,
        device_id: Optional[str] = None,
        external_id: Optional[str] = None,
    ) -> List[User]:
        if user_uuid:
            return self._client.run(
                self._get_user(
                    user_uuid=user_uuid,
                )
            )

        return self._client.run(
            self._get(
                created_after=created_after,
                created_before=created_before,
                search=search,
                phone_number=phone_number,
                device_id=device_id,
                external_id=external_id,
            )
        )



from abc import ABC
from typing import Any, Dict, List, TypedDict

from pydantic import validate_arguments

from ..cache import cache
from ..core import BaseClient
from ..errors import RequestError, ServerError

PATH = "/users/{user_uuid}/institutions"

Institution = Dict[str, Any]


class InstitutionsResponse(TypedDict):
    institutions: List[Institution]


class BaseInstitutionsResource(ABC):
    def __init__(self, client: BaseClient):
        self._client = client

    @validate_arguments
    @cache
    async def _get(self, user_uuid: str) -> List[Institution]:
        async with self._client.session() as session:
            response = await session.get(PATH.format(user_uuid=user_uuid))

        if response.status_code // 100 == 5:
            raise ServerError(response.text)

        if response.status_code // 100 == 4:
            raise RequestError(response.text)

        institutions_response: InstitutionsResponse = response.json()
        return institutions_response["institutions"]


class AsyncInstitutionsResource(BaseInstitutionsResource):
    async def get(self, user_uuid: str) -> List[Institution]:
        return await self._get(user_uuid)


class SyncInstitutionsResource(BaseInstitutionsResource):
    def get(self, user_uuid: str) -> List[Institution]:
        return self._client.run(self._get(user_uuid))



from .core import BaseClient
from .resources.alerts import AsyncAlertsResource, SyncAlertsResource
from .resources.balances import AsyncBalancesResource, SyncBalancesResource
from .resources.institutions import AsyncInstitutionsResource, SyncInstitutionsResource
from .resources.transactions import AsyncTransactionsResource, SyncTransactionsResource
from .resources.users import AsyncUsersResource, SyncUsersResource


class AsyncClient(BaseClient):
    def __init__(
        self,
        access_token: str,
        concurrency_limit: int = 50,
        base_url: str = "https://api.pngme.com/beta",
    ):
        super().__init__(
            access_token=access_token,
            concurrency_limit=concurrency_limit,
            base_url=base_url,
        )

        self.alerts = AsyncAlertsResource(self)
        self.balances = AsyncBalancesResource(self)
        self.institutions = AsyncInstitutionsResource(self)
        self.transactions = AsyncTransactionsResource(self)
        self.users = AsyncUsersResource(self)


class Client(BaseClient):
    def __init__(
        self,
        access_token: str,
        concurrency_limit: int = 50,
        base_url: str = "https://api.pngme.com/beta",
    ):
        super().__init__(
            access_token=access_token,
            concurrency_limit=concurrency_limit,
            base_url=base_url,
        )

        self.alerts = SyncAlertsResource(self)
        self.balances = SyncBalancesResource(self)
        self.institutions = SyncInstitutionsResource(self)
        self.transactions = SyncTransactionsResource(self)
        self.users = SyncUsersResource(self)
